#include <stdio.h>
#include <string.h>
#include <stddef.h>
#include <mpi.h>
#include <stdlib.h>

#include <dirent.h>

#include "tpl.h"

#include "sequence.h"
#include "d7.h"

#include "queue.h"

#define ITER 3
#define WITER 1

#define BALANCE_MAX_CHUNK 32	// how many d7s may be balanced per communication?
#define BALANCE_UPPER 0.7		// threshold for worker to request balance
#define BALANCE_LOWER 0.3		// threshold for worker to request work

#define NLEN 200					// debug namelen

#define DEBUG_DISPLAY_COMM 0
#define DEBUG_DISPLAY_STATUS 0
#define DEBUG_DISPLAY_STATS 1

// MPI related variables
MPI_Status status;
MPI_Request request;
int numprocs;
int myrank;

// MPI + Queue 
MPI_Datatype mpi_d7;

// Queue related variables
int iter = ITER;
int witer = WITER;


int * qm;			// queue monitor, holds information how many d7s are enqueued
					// on a given node qm[rank]

int * qm_demand;	// cache demands



// temporary to keep testing simple
struct d7 r;		// receive single d7 buffer
struct d7 s;		// send single d7 buffer

void print_queue()
{
	char seq_name[NLEN+1];
	seq_name[NLEN] = (char)0;
	
	for (int t = 0; t < lifo_count; ++t)
	{
		for (int i = 0; i < NLEN; ++i)
		{
			seq_name[i] = '0' + (char)lifo[t].sequence[i];
		}

		printf("W %d, %d | lifo[%d]=%s turn=%d\n", myrank, iter, t, seq_name, lifo[t].turn );
	}
}


void dump_d7(d7 game)
{
	tpl_node *tn;

	char seq_name[NLEN+1];
	seq_name[NLEN]=(char)0;
	
	char filename[255];

	for (int i = 0; i < NLEN; ++i)
	{
		seq_name[i] = '0' + (char)game.sequence[i];
	}

	sprintf(filename, "queue/%s.tpl", seq_name);

	tn = tpl_map("S(i##i#iiiiiii)", &game, 7, 7, 3750);  /* pass structure address */
	tpl_pack(tn, 0);
	tpl_dump(tn, TPL_FILE, filename);
	tpl_free(tn);

}

void dump_queue()
{
	char seq_name[NLEN+1];
	seq_name[NLEN] = (char)0;
	
	for (int t = 0; t < lifo_count; ++t)
	{
		dump_d7(lifo[t]);
	}
}



void gameover(d7 game)
{
	FILE *fp;
	char filename[255];
	
	
	char seq_name[NLEN+1];
	seq_name[NLEN]=(char)0;
	for (int i = 0; i < NLEN; ++i)
	{
		seq_name[i] = '0' + (char)game.sequence[i];
	}

	sprintf(filename, "gameover/%s.d7", seq_name);

	fp = fopen(filename, "w");
	if (fp == NULL) 
	{
		printf("I couldn't open %s for writing.\n", filename);
	} else {
		printf("GAME OVER! Scored %d with %s.\n", game.score, seq_name);
	}

	fprintf(fp, "score: %d\nsequence:\n%s", game.score, seq_name);
	fclose(fp);
}

void calculate(d7 game)
{
	d7 ngame;

	char seq_name[NLEN+1];
	seq_name[NLEN]=(char)0;
	for (int i = 0; i < NLEN; ++i)
	{
		seq_name[i] = '0' + (char)game.sequence[i];
	}
	//printf("W %d, %d | processing game=%s \n", myrank, iter, seq_name );

	for (int pos = 0; pos < 7; ++pos)
	{
		ngame = game;
		ngame.status = D7_STATUS_REQUIRE_DISC;
		ngame = dropdisc(ngame, pos, sequence[ngame.turn]);

		if ( ngame.status != D7_STATUS_REQUIRE_DISC )
		{
			ngame.sequence[ngame.turn] = pos;
			ngame.turn++;
			ngame.discs--;

			ngame = calcturn(ngame);

			if ( ngame.discs == 0 ) {
				ngame.level++;
				ngame = levelup(ngame, rows[ngame.level]);

				// calc turn again after levelup as ngame board changed
				ngame = calcturn(ngame);
			}

			// reset depth counter as turn is over
			ngame.depth = 0;

			if ( ngame.status == D7_STATUS_GAME_OVER ) {
				gameover(ngame);
			} else {
				push(ngame);
			}
		}

	}

}



double qm_mean()
{
	double mean = 0;
	for (int p = 0; p < numprocs; ++p)
	{
		mean += qm[p];
	}
	return mean/numprocs;
}

double qm_wmean()
{
	double mean = 0;
	for (int p = 1; p < numprocs; ++p)
	{
		mean += qm[p];
	}
	return mean/(numprocs-1);
}


void qm_print()
{
	qm[myrank] = lifo_count;
	
	printf("M %d | ", myrank);
	for (int p = 0; p < numprocs; ++p)
	{
		printf("qm[%d]: %d, ", p, qm[p]);
	}

	printf("\n    | mean: %f, wmean: %f\n\n", qm_mean(), qm_wmean());

}


void qm_gather()
{
	for ( int p = 1; p < numprocs; p++ )
	{	
		// get queue status
		MPI_Recv( &qm[p], 1, MPI_INT, p, 1, MPI_COMM_WORLD, &status );

		// if current workers queue is not empty it also sends a sample
		if ( qm[p] > 0 )
		{	
			printf("M %d, %d | %d | worker%d is OK \n", myrank, iter, p, p );
			//printf("M %d, %d | %d | rank: %d id: %d, text: %s \n", myrank, iter, p, r.rank, r.id, r.text);
			//printf("M %d, %d | %d | lifo_count=%d \n", myrank, iter, p, lifo_count);

		} else {
			printf("M %d, %d | %d | worker%d ran out of jobs! \n", myrank, iter, p, p );
		}

		printf("-------+---+-------------------------------------\n");
	}

	qm[myrank] = lifo_count;
}


void qm_report()
{
	// inform master about queue length
	MPI_Isend( &lifo_count , 1, MPI_INT, 0, 1, MPI_COMM_WORLD, &request );
	// send a sample/ or not
	if ( lifo_count > 0 )
	{
		// read a sample and send back to master
		//s = top(); 
		// send sample to master (sample is not removed from queue!)
		//MPI_Isend( &s , 1, mpi_d7, 0, 1, MPI_COMM_WORLD, &request );
	} else {
		// balance could happen here
	}

}


void qsend(int num, int rank)
{
	struct d7 * d7_buf; 
	d7_buf = (d7*)malloc( num * sizeof(d7));	

	char seq_name[NLEN+1];
	seq_name[NLEN]=(char)0;

	for (int d = 0; d < num; ++d)
	{
		d7_buf[d] = pop();

		for (int i = 0; i < NLEN; ++i)
		{
			seq_name[i] = '0' + (char)d7_buf[d].sequence[i];
		}

		if ( DEBUG_DISPLAY_COMM ) {
			printf("W %d, %d | for %d > packing  game=%s \n", myrank, iter, rank, seq_name);
			printf("W %d, %d | from %d > m6,0=%d score=%d level=%d nld=%d turn=%d discs=%d status=%d depth=%d\n", myrank, iter, rank, d7_buf[d].m[6][0], d7_buf[d].score, d7_buf[d].level, d7_buf[d].nextleveldiscs, d7_buf[d].turn, d7_buf[d].discs, d7_buf[d].status, d7_buf[d].depth);
		}
		//printf("W %d, %d | for %d: d7.text=%s \n", myrank, iter, rank, d7_buf[d].text);
	}


	if ( DEBUG_DISPLAY_COMM ) {
		printf("W %d, %d | sending %d d7(s) to %d \n", myrank, iter, num, rank);
	}
	MPI_Send( d7_buf , num, mpi_d7, rank, 1, MPI_COMM_WORLD );
	
	free(d7_buf);
}

void qrecv(int num, int rank)
{
	struct d7 * d7_buf;
	d7_buf = (d7*)malloc( num * sizeof(d7));	

	char seq_name[NLEN+1];
	seq_name[NLEN]=(char)0;
	
	MPI_Recv( d7_buf, num, mpi_d7, rank, 1, MPI_COMM_WORLD, &status );
	
	if ( DEBUG_DISPLAY_COMM ) {
		printf("W %d, %d | recieved %d d7(s) from %d\n", myrank, iter, num, rank);
	}

	for (int d = 0; d < num; ++d)
	{
		for (int i = 0; i < NLEN; ++i)
		{
			seq_name[i] = '0' + (char)d7_buf[d].sequence[i];
		}
	

		if ( DEBUG_DISPLAY_COMM )
		{
			printf("W %d, %d | from %d recieved game=%s \n", myrank, iter, rank, seq_name );
			printf("W %d, %d | from %d > m6,0=%d score=%d level=%d nld=%d turn=%d discs=%d status=%d depth=%d\n", myrank, iter, rank, d7_buf[d].m[6][0], d7_buf[d].score, d7_buf[d].level, d7_buf[d].nextleveldiscs, d7_buf[d].turn, d7_buf[d].discs, d7_buf[d].status, d7_buf[d].depth);
		}

		push(d7_buf[d]);
	}

	free(d7_buf);
}

int sanitize_demand(int n)
{
	if ( n > BALANCE_MAX_CHUNK )
	{
		n = BALANCE_MAX_CHUNK;
	}	
	return n;
}



void work() 
{
	while ( witer > 0)
	{
		if ( lifo_count > 0 )
		{
			calculate(pop());
		} else {
			//TODO: request work
			witer = 0;
			break;
		}

		witer--;
	}
}

void test_queuefolder()
{
	int file_count = 0;
	DIR * dirp;
	struct dirent * entry;

	dirp = opendir("queue"); /* There should be error handling after this */
	while ((entry = readdir(dirp)) != NULL) {
		if (entry->d_type == DT_REG) { /* If the entry is a regular file */
				 file_count++;
			}
	}
	closedir(dirp);

	printf("Files in ./queue/: %d\n", file_count);
}


int main( int argc, char *argv[] )
{
	// easy access to environment variables (e.g. numprocs, rank, ...)
    int namelen, i;
    char processor_name[MPI_MAX_PROCESSOR_NAME];

	// prepare costum datatype
    MPI_Datatype types[] = {
		MPI_INT, // m 
		MPI_INT, // sequence
		MPI_INT, // score
		MPI_INT, // level
		MPI_INT, // nextleveldiscs
		MPI_INT, // turn
		MPI_INT, // discs
		MPI_INT, // status
		MPI_INT, // depth
	};

	int blocklengths[] = { 
		7*7,  // m 
		3750, // sequence
		1,    // score
		1,    // level
		1,    // nextleveldiscs
		1,    // turn
		1,    // discs
		1,    // status
		1,    // depth
	};
	
	MPI_Aint displacements[] = {
		offsetof(d7, m),			// m 
		offsetof(d7, sequence),		// sequence
		offsetof(d7, score),		// score
		offsetof(d7, level),		// level
		offsetof(d7, nextleveldiscs),	// nextleveldiscs
		offsetof(d7, turn),			// turn
		offsetof(d7, discs),		// discs
		offsetof(d7, status),		// status
		offsetof(d7, depth),		// depth
	};

	// init mpi
    MPI_Init( &argc, &argv );

	// create costum datatype
	MPI_Type_create_struct(
			9,  // double-check to match number of struct elements  ;)
			blocklengths, displacements, types,
			&mpi_d7
	);
    MPI_Type_commit(&mpi_d7);

	// populate some environment variables
    MPI_Comm_size( MPI_COMM_WORLD, &numprocs );
    MPI_Comm_rank( MPI_COMM_WORLD, &myrank );
    MPI_Get_processor_name( processor_name, &namelen );

	// init lifo for x elements, may increase 
	lifo = (d7*)malloc(sizeof(d7) * 10);
	lifo_count = -1;
	lifo_len = 10;

	// allocate array to store lifo_count of every node
	qm = malloc(sizeof(int) * numprocs );
	qm_demand = malloc(sizeof(int) * numprocs );



	// some temporary inits
	int rint;
	int rsize;
	int demand;

	d7 game;

    if ( myrank == 0 )
	{
		/*
		 * Master
		 */


		// init master
		//
		
		test_queuefolder();
		
		// init game
		d7 game = {{{0}}, {9}, 0, 0, 30, 0, 30, D7_STATUS_START_GAME, 0};
		game = levelup(game, rows[game.level] );
	    push(game);


		while ( lifo_count < numprocs && lifo_count > 0 ) 
		{
			// not enough elements in queue, get more
			// queue not empty?
			calculate(pop());
		}

		printf("M %d | lifo_count = %d \n", myrank, lifo_count);

		// share the d7s
		if ( lifo_count > 0 )
		{
			for ( i = 1; i < numprocs; i++ )
			{	
				qsend(1, i);
				//s = pop();
				//MPI_Send( &s, 1, mpi_d7, i, 1, MPI_COMM_WORLD );
			}	
		}

		// handle queue
		printf("M %d, %d: lifo_count=%d \n", myrank, iter, lifo_count);
		printf("M %d, %d: starting while \n\n", myrank, iter);
	


		//// begin master normal
		while ( iter > 0 )
		{

			// update qm stats
			qm_gather();
			qm_print();

			// Broadcast demands
			for ( i = 1; i < numprocs; i++ )
			{	
				demand = qm[i] - (int)qm_wmean();
				demand = sanitize_demand(demand);
				qm_demand[i] = demand;
				//printf("M %d | %d | from %d demanded %d \n", myrank, iter, i, demand);
				MPI_Isend( &demand , 1, MPI_INT, i, 1, MPI_COMM_WORLD, &request );
			}
		
			// Collect demands
			printf("\nM %d, %d | Collecting demands.\n", myrank, iter);
			for ( i = 1; i < numprocs; i++ )
			{
				demand = qm_demand[i];
				if ( demand > 0 )
				{
					qrecv(demand, i);
				}
			}
			printf("M %d, %d | Done collecting demands.\n\n", myrank, iter);

			qm_print();

			// Fill up lazy workers
			printf("\nM %d, %d | Fill up lazy workers.\n", myrank, iter);
			for ( i = 1; i < numprocs; i++ )
			{
				demand = qm_demand[i];
				demand = demand*(-1);
				demand = sanitize_demand(demand);
				if ( demand > 0 )
				{
					qsend(demand, i);
				}
			}
			printf("M %d, %d | Done filling up lazy workers.\n\n", myrank, iter);



			//printf("M %d, %d | finished a normal\n", myrank, iter);


			iter--;
		} //// end master normal


		// final qather just before exit
		// update qm stats
		qm_gather();
		qm_print();

		dump_queue();

	}
    else
	{
		/*
		 * Worker
		 */

		//// begin worker init
		// get initial work seed		
		qrecv(1, 0);
		//// end worker init



		//// begin worker normal
		while ( iter > 0 )
		{
			// workiter
			witer = WITER;

			// work
			work();


			// report queue status to master
			qm_report();
		

			// recv demand
			MPI_Recv( &demand, 1, MPI_INT, 0, 1, MPI_COMM_WORLD, &status );
			// fullfill demand
			if ( demand > 0 )
			{
				qsend(demand, 0);
			} else if (demand < 0) {
				// TODO: recv d7s?
				demand = demand*(-1);
				demand = sanitize_demand(demand);
				qrecv(demand,0);
			}
			
			//printf("W %d, %d | finished a normal\n", myrank, iter);


			// another cycle finished
			iter--;

		} //// end worker normal


		// final report just before exit
		qm_report();

    }

	// barrier to prevent outstanding communications from creating deadlocks
	printf("N %d | waiting before final B\n",myrank);
	MPI_Barrier(MPI_COMM_WORLD);
	printf("N %d | passed final B\n",myrank);

	// clean up to prevent memory leaks
	free(lifo);
	free(qm);

	// got here? nicely done, Drop7 solved.
	MPI_Finalize();
    return 0;
}
